@@ -1,6 +1,7 @@
 import logging
 import sys
 import time
+import json
 from typing import Any, Callable, Dict, List, Optional, Union, Tuple
 
 import grpc
@@ -29,9 +30,15 @@ def validate_type(
     given,  # type: Any
     expected_types,  # type: List[type]
 ):
-    if type(given) not in expected_types:
+    if not any([isinstance(given, t) for t in expected_types]):
+        method_name = sys._getframe(1).f_code.co_name
+
+        # e.g.
+        # post_text input 'None' has type <class 'NoneType'>,
+        # but expected one of: [<class 'str'>]
         raise TypeError(
-            "Formant client method input '"
+            method_name
+            + " input '"
             + str(given)
             + "' has type "
             + str(type(given))
@@ -103,6 +110,9 @@ class Client:
         self._telemetry_listener_callback_streams = (
             {}
         )  # type: Dict[Callable[..., None], CancellableStreamThread]
+        self._custom_data_channel_message_callback_streams = (
+            {}
+        )  # type: Dict[Callable[..., None], CancellableStreamThread]
 
         self._setup_agent_communication()
 
@@ -123,7 +133,7 @@ class Client:
         :param stream: The name of the Formant stream to post the datapoint on
         :param value: The text datapoint value
         :param tags: Tags to include on the posted datapoint
-        :param timestamp: Unix timestamp for the posted datapoint.
+        :param timestamp: Unix timestamp in milliseconds for the posted datapoint.
             Uses the current time by default
 
         .. highlight:: python
@@ -164,7 +174,7 @@ class Client:
         :param stream: The name of the Formant stream to post the datapoint on
         :param value: The encoded JSON datapoint value
         :param tags: Tags to include on the posted datapoint
-        :param timestamp: Unix timestamp for the posted datapoint.
+        :param timestamp: Unix timestamp in milliseconds for the posted datapoint.
             Uses the current time by default
         """
 
@@ -194,7 +204,7 @@ class Client:
         :param stream: The name of the Formant stream to post the datapoint on
         :param value: The numeric datapoint value
         :param tags: Tags to include on the posted datapoint
-        :param timestamp: Unix timestamp for the posted datapoint.
+        :param timestamp: Unix timestamp in milliseconds for the posted datapoint.
             Uses the current time by default
         """
 
@@ -226,7 +236,7 @@ class Client:
         :param numerics_dict: The numeric set datapoint value,
             a dictionary mapping names to (numeric value, units) tuples.
         :param tags: Tags to include on the posted datapoint
-        :param timestamp: Unix timestamp for the posted datapoint.
+        :param timestamp: Unix timestamp in milliseconds for the posted datapoint.
             Uses the current time by default
 
         .. highlight:: python
@@ -251,6 +261,10 @@ class Client:
         for k, v in numerics_dict.items():
             validate_type(k, [str])
             validate_type(v, [tuple])
+            if len(v) != 2:
+                raise TypeError(
+                    "value %s for key %s in numericset must have length of 2" % (v, k)
+                )
             validate_type(v[0], [float, int])
             if v[1]:
                 validate_type(v[1], [str])
@@ -291,7 +305,7 @@ class Client:
         :param content_type: The format of the encoded image or frame.
             Defaults to "image/jpg".
         :param tags: Tags to include on the posted datapoint
-        :param timestamp: Unix timestamp for the posted datapoint.
+        :param timestamp: Unix timestamp in milliseconds for the posted datapoint.
             Uses the current time by default
         """
         if value is not None and url is not None:
@@ -326,7 +340,7 @@ class Client:
         :param stream: The name of the Formant stream to post the datapoint on
         :param bitset_dict: The datapoint value, a dictionary mapping names to booleans
         :param tags: Tags to include on the posted datapoint
-        :param timestamp: Unix timestamp for the posted datapoint.
+        :param timestamp: Unix timestamp in milliseconds for the posted datapoint.
             Uses the current time by default
 
         .. highlight:: python
@@ -370,6 +384,8 @@ class Client:
         longitude,  # type: Union[float, int]
         tags=None,  # type: Optional[Dict[str, str]]
         timestamp=None,  # type: Optional[int]
+        altitude=None,  # type: Union[float, int]
+        orientation=None,  # type: Union[float, int]
     ):
         # type: (...) -> None
         """
@@ -379,7 +395,7 @@ class Client:
         :param latitude: The datapoint value's latitude
         :param longitude: The datapoint value's longitude
         :param tags: Tags to include on the posted datapoint
-        :param timestamp: Unix timestamp for the posted datapoint.
+        :param timestamp: Unix timestamp in milliseconds for the posted datapoint.
             Uses the current time by default
         """
 
@@ -390,7 +406,8 @@ class Client:
             datapoint_pb2.Datapoint(
                 stream=stream,
                 location=navigation_pb2.Location(
-                    latitude=latitude, longitude=longitude
+                    latitude=latitude, longitude=longitude, altitude=altitude,
+                    orientation=orientation,
                 ),
                 tags=tags if tags else {},
                 timestamp=timestamp if timestamp else current_timestamp(),
@@ -418,7 +435,7 @@ class Client:
         :param current: The battery current
         :param charge: The battery charge
         :param tags: Tags to include on the posted datapoint
-        :param timestamp: Unix timestamp for the posted datapoint.
+        :param timestamp: Unix timestamp in milliseconds for the posted datapoint.
             Uses the current time by default
         """
 
@@ -654,7 +671,7 @@ class Client:
         :param f: A callback that will be executed on command requests
             as they are received by the Formant agent.
         :param command_filter: A list of command names. The provided callback
-            is only exectued on commands whose names are in this list
+            is only executed on commands whose names are in this list
         """
 
         def create_stream():
@@ -767,6 +784,193 @@ class Client:
         if stream is not None:
             stream.cancel()
 
+    # Custom data channel methods
+
+    @handle_agent_exceptions
+    def send_on_custom_data_channel(
+        self,
+        channel_name,  # type: str
+        payload,  # type: bytes
+    ):
+        self.agent_stub.SendOnCustomDataChannel(
+            agent_pb2.SendOnCustomDataChannelRequest(
+                channel_name=channel_name, payload=payload
+            )
+        )
+
+    def register_custom_data_channel_message_callback(
+        self,
+        f,  # type: Callable[[agent_pb2.GetCustomDataChannelMessageStreamResponse], None] # noqa: E501
+        channel_name_filter=None,  # type: Optional[List[str]]
+    ):
+        # type : (...) -> None
+        """
+        :param f: A callback that will be called with messages
+            received on the specified custom data channel
+        :param channel_name_filter: An optional allow list of custom channel names
+            for this callback
+        """
+
+        def create_stream():
+            return self._get_custom_data_channel_message_stream(
+                channel_name_filter=channel_name_filter
+            )
+
+        self._custom_data_channel_message_callback_streams[f] = CancellableStreamThread(
+            f, create_stream, self.logger, ignore_unavailable=self.ignore_unavailable
+        )
+
+    def unregister_custom_data_channel_message_callback(
+        self, f,  # type: Callable[[], None]
+    ):
+        # type : (...) -> None
+        """
+        :param f: The custom data channel message callback to be unregistered
+        """
+        stream = self._custom_data_channel_message_callback_streams.pop(f, None)
+        if stream is not None:
+            stream.cancel()
+
+    def custom_data_channel_request_handler(
+        self, channel_name  # type: str
+    ):
+        # type : (...) -> Callable[Callable[agent_pb2.GetCustomDataChannelMessageStreamResponse, None], Callable[str, str]]: # noqa: E501
+        """
+        Registers a handler for requests sent by RequestDataChannel instances
+        (part of the Formant toolkit).
+        See: https://github.com/FormantIO/toolkit/tree/master/examples/request-response
+        for an example.
+
+        :param channel_name: The name of the custom data channel to listen on
+
+        .. highlight:: python
+        .. code-block:: python
+
+            from formant.sdk.agent.v1 import Client
+
+            fclient = Client()
+
+            @fclient.custom_data_channel_request_handler("my_channel")
+            def handler(request_data):
+                # Do something with request_data string
+                print(json.loads(request_data))
+
+                # Return any string response
+                return json.dumps({"message": "Hello world!"})
+        """
+
+        def outer_wrapper(f):
+            def inner_wrapper(message):
+                try:
+                    payload = json.loads(message.payload.decode("utf-8"))
+                    request_id = payload["id"]
+                    request_data = payload["data"]
+                except (
+                    UnicodeDecodeError,
+                    json.decoder.JSONDecodeError,
+                    KeyError,
+                ) as e:
+                    if self.logger:
+                        self.logger.error(
+                            "Received invalid custom data channel request: "
+                            + e.__class__.__name__
+                            + ": "
+                            + str(e)
+                        )
+
+                response_data = None
+                try:
+                    response = f(request_data)
+                    # Place valid responses on the "data" key
+                    response_data = json.dumps(
+                        {"id": request_id, "data": response}
+                    ).encode("utf-8")
+                except Exception as e:
+                    error_message = (
+                        "An error occurred while handling custom data channel request: "
+                        + e.__class__.__name__
+                        + ": "
+                        + str(e)
+                    )
+                    # Place error responses on the "error" key
+                    response_data = json.dumps(
+                        {"id": request_id, "error": error_message}
+                    ).encode("utf-8")
+                self.send_on_custom_data_channel(channel_name, response_data)
+
+            # When the handler function is decorated,
+            # we automatically register it as a callback
+            self.register_custom_data_channel_message_callback(
+                inner_wrapper, [channel_name]
+            )
+
+        return outer_wrapper
+
+    def custom_data_channel_binary_request_handler(
+        self, channel_name  # type: bytes
+    ):
+        # type : (...) -> Callable[Callable[agent_pb2.GetCustomDataChannelMessageStreamResponse, None], Callable[str, str]]: # noqa: E501
+        """
+        :param channel_name: The name of the custom data channel to listen on
+
+        .. highlight:: python
+        .. code-block:: python
+
+            from formant.sdk.agent.v1 import Client
+
+            fclient = Client()
+
+            @fclient.custom_data_channel_request_handler("my_channel")
+            def handler(request_data):
+                # Do something with request_data bytes
+                print(request_data.decode("utf-8"))
+
+                # Return any bytes response
+                return b"Hello."
+        """
+
+        SUCCESS_RESPONSE_BYTE = b"\00"
+        ERROR_RESPONSE_BYTE = b"\01"
+
+        def outer_wrapper(f):
+            def inner_wrapper(message):
+                try:
+                    request_id = message.payload[0:16]
+                    request_data = message.payload[16:]
+                except TypeError as e:
+                    if self.logger:
+                        self.logger.error(
+                            "Received invalid custom data channel request: "
+                            + e.__class__.__name__
+                            + ": "
+                            + str(e)
+                        )
+
+                response_data = None
+                try:
+                    response = f(request_data)
+                    response_data = request_id + SUCCESS_RESPONSE_BYTE + response
+                except Exception as e:
+                    error_message = (
+                        "An error occurred while handling custom data channel request: "
+                        + e.__class__.__name__
+                        + ": "
+                        + str(e)
+                    )
+                    response_data = (
+                        request_id + ERROR_RESPONSE_BYTE + error_message.encode("utf-8")
+                    )
+
+                self.send_on_custom_data_channel(channel_name, response_data)
+
+            # When the handler function is decorated,
+            # we automatically register it as a callback
+            self.register_custom_data_channel_message_callback(
+                inner_wrapper, [channel_name]
+            )
+
+        return outer_wrapper
+
     # Reading application configuration
 
     def get_app_config(self, key, *args):
@@ -865,3 +1069,10 @@ class Client:
         return self.agent_stub.GetTelemetryListenerStream(
             agent_pb2.GetTelemetryListenerStreamRequest(stream_filter=stream_filter)
         )
+
+    def _get_custom_data_channel_message_stream(self, channel_name_filter=None):
+        return self.agent_stub.GetCustomDataChannelMessageStream(
+            agent_pb2.GetCustomDataChannelMessageStreamRequest(
+                channel_name_filter=channel_name_filter
+            )
+        )
